{% extends "!footer.html" %}

{% block extrafooter %}
<br>
<div>
  <p>
      For more complete information about compiler optimizations, see our <a href="{{ pathto('optimization_notice') }}">Optimization Notice</a>.
  </p>  
</div>
{% endblock %}