{% extends "!layout.html" %}
{% block extrahead %}

    <script defer type="text/javascript" src="https://www.intel.com/content/dam/www/global/wap/performance-config.js" ></script>
    <script type="text/javascript">
        // Configure TMS settings
        var wapLocalCode = 'us-en'; // Dynamically set per localized site, see mapping table for values
        var wapSection = "oneapi-dpl"; // WAP team will give you a unique section for your site
        // Load TMS
        if(document.location.href.includes("uxlfoundation.github.io")){
        (function () {
        var url = 'https://www.intel.com/content/dam/www/global/wap/tms-loader.js'; // WAP file URL
        var po = document.createElement('script'); po.type = 'text/javascript'; po.async = true; po.src = url;
        var s = document.getElementsByTagName('script')[0]; s.parentNode.insertBefore(po, s);
        })();
        }
    </script>
    <link href="{{ pathto("_static/style.css", True) }}" rel="stylesheet" type="text/css">


{% endblock %}

          {% block sidebartitle %}

          {% if logo and theme_logo_only %}
            <a href="{{ pathto(master_doc) }}">
          {% else %}
            <a href="https://docs.oneapi.com/index.html" class="icon icon-home"> {{ project }}
          {% endif %}

          {% if logo %}
            {# Not strictly valid HTML, but it's the only way to display/scale
               it properly, without weird scripting or heaps of work
            #}
            <img src="{{ pathto('_static/' + logo, 1) }}" class="logo" alt="Logo"/>
          {% endif %}
          </a>

          {% if theme_display_version %}
            {%- set nav_version = version %}
            {% if READTHEDOCS and current_version %}
              {%- set nav_version = current_version %}
            {% endif %}
            {% if nav_version %}
              <div class="version">
                {{ nav_version }}
              </div>
            {% endif %}
          {% endif %}

          {% include "searchbox.html" %}

          {% endblock %}
