//==============================================================
// Copyright © 2020 Intel Corporation
//
// SPDX-License-Identifier: MIT
// =============================================================

#ifndef _GAMMA_UTILS_HPP
#define _GAMMA_UTILS_HPP

#include "utils/Img.hpp"
#include "utils/ImgAlgorithm.hpp"
#include "utils/ImgFormat.hpp"
#include "utils/ImgPixel.hpp"

#include "utils/Other.hpp"

#endif  // _GAMMA_UTILS_HPP
